/* Copyright (c) 1991, 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* Lisp-style data management routines. */

/* (This should have some deallocation support, since some game init data
   can be discarded (but how much?) */

#include "config.h"
#include "misc.h"
#include "lisp.h"

/* The only game.h thing we need. */

#define NONUTYPE (MAXUTYPES)
#define NONMTYPE (MAXMTYPES)
#define NONTTYPE (MAXTTYPES)

extern short numutypes, nummtypes, numttypes;

/* Pointer to "nil", the empty list. */

Obj *lispnil;

/* Pointer to "eof", which is returned if no more forms in a file. */

Obj *lispeof;

/* Symbol table variables.  The table itself is dynamically allocated. */

int numsymbols = 0;

/* The topend on symbols. */

int maxsymbols = 1000;

Symentry *symboltable = NULL;

/* The count of the number of Lisp objects allocated. */

long lispmalloc = 0;

/* This variable is used to track the depth of nested #| |# comments. */

int commentdepth = 0;

/* Allocate a new Lisp object, count it as such. */

Obj *
newobj()
{
    lispmalloc += sizeof(Obj);
    return (Obj *) xmalloc(sizeof(Obj));
}

/* Pre-create some objects that should always exist. */

init_lisp()
{
    Obj *randsym;

    lispnil = newobj();
    lispnil->type = NIL;
    /* So car/cdr of nil is nil, possibly makes infinite loops tho */
    lispnil->v.cons.car = lispnil;
    lispnil->v.cons.cdr = lispnil;
    /* We use the eof object to recognize eof when reading a file. */
    lispeof = newobj();
    lispeof->type = EOFOBJ;
    /* Set up the symbol table. */
    symboltable = (Symentry *) xmalloc(maxsymbols * sizeof(Symentry));
    numsymbols = 0;
    /* Make some predefined symbols. */
    setq(intern_symbol("false"), new_number(0));
    setq(intern_symbol("true"), new_number(1));
    setq(intern_symbol("non-unit"), new_number(NONUTYPE));
    setq(intern_symbol("non-material"), new_number(NONMTYPE));
    setq(intern_symbol("non-terrain"), new_number(NONTTYPE));
    setq(intern_symbol("open"), new_number(0));
    setq(intern_symbol("border"), new_number(1));
    setq(intern_symbol("connection"), new_number(2));
    setq(intern_symbol("coating"), new_number(3));
    /* Leave these unbound so that first ref computes correct list. */
    intern_symbol("u*");
    intern_symbol("m*");
    intern_symbol("t*");
    randsym = intern_symbol("and");
    setq(randsym, randsym);
    randsym = intern_symbol("or");
    setq(randsym, randsym);
    randsym = intern_symbol("not");
    setq(randsym, randsym);
}

/* El cheapo Lisp reader.  Lisp objects are generally advertised by their
   first characters, but lots of semantics actions happen while reading, so
   this isn't really a regular expression reader. */

#define closingparen 45

#define BIGBUF 1000

int *startlineno;
int *endlineno;

Obj *
read_form(fp, p1, p2)
FILE *fp;
int *p1, *p2;
{
    char ch;

    commentdepth = 0;
    startlineno = p1;
    endlineno = p2;
    return read_form_aux(fp);
/*	    init_warning("extra closing paren, ignoring"); */
}

Obj *
read_form_aux(fp)
FILE *fp;
{
    int minus, factor, commentclosed;
    int ch, ch2, ch3, ch4;
    char str[BIGBUF];
    int j = 0, num;

    while ((ch = getc(fp)) != EOF) {
	/* Recognize nested comments specially. */
	if (ch == '#') {
	    if ((ch2 = getc(fp)) == '|') {
		commentclosed = FALSE;
		++commentdepth;
		while ((ch3 = getc(fp)) != EOF) {
		    if (ch3 == '|') {
			/* try to recognize # */
			if ((ch4 = getc(fp)) == '#') {
			    --commentdepth;
			    if (commentdepth == 0) {
				commentclosed = TRUE;
				break;
			    }
			} else {
			    ungetc(ch4, fp);
			}
		    } else if (ch3 == '#') {
			if ((ch4 = getc(fp)) == '|') {
			    ++commentdepth;
			} else {
			    ungetc(ch4, fp);
			}
		    } else if (ch3 == '\n') {
			++(*endlineno);
			announce_read_progress();
		    }
		}
		if (!commentclosed) {
		    init_warning("comment not closed at eof");
		}
		/* Always pick up the next char. */
		ch = getc(fp);
	    } else {
		ungetc(ch2, fp);
	    	return intern_symbol("#");
	    }
	}
	/* Regular lexical recognition. */
	if (iswhite(ch)) {
	    /* Nothing to do here except count lines. */
	    if (ch == '\n') {
		++(*endlineno);
		++(*startlineno);
		announce_read_progress();
	    }
	} else if (ch == ';') {
	    /* Discard all from here to the end of this line. */
	    while ((ch = getc(fp)) != EOF && ch != '\n');
	    ++(*endlineno);
	    announce_read_progress();
	} else if (ch == '(') {
	    return read_list(fp, startlineno, endlineno);
	} else if (ch == ')') {
	    return (Obj *) closingparen; /* yech */
	} else if (ch == '"') {
	    /* handle escape chars? */
	    while ((ch = getc(fp)) != EOF && ch != '"')
	      if (j < BIGBUF) str[j++] = ch;
	    str[j] = '\0';
	    return new_string(copy_string(str));
	} else if (ch == '|') {
	    while ((ch = getc(fp)) != EOF && ch != '|')
	      if (j < BIGBUF) str[j++] = ch;
	    str[j] = '\0';
	    return intern_symbol(str);
	} else if (isdigit(ch) || ch == '-' || ch == '+' || ch == '.') {
	    minus = (ch == '-');
	    factor = (ch == '.' ? 100 : 1);
	    num = (minus ? 0 : ch - '0');
	    while ((ch = getc(fp)) != EOF) {
	    	if (isdigit(ch)) {
	    	    /* should ignore decimal digits past second one */
		    num = num * 10 + ch - '0';
		    if (factor > 1) factor /= 10;
		} else if (ch == 'd') {
		} else if (ch == '+') {
		} else if (ch == '.') {
		    factor = 100;
		} else {
		    break;
		}
	    }
	    ungetc(ch, fp);
	    if (minus) num = 0 - num;
	    return new_number(factor * num);
	} else { 
	    str[j++] = ch;
	    while ((ch = getc(fp)) != EOF &&
		   !iswhite(ch) &&
		   !strchr("();\"'`#", ch)) {
		if (j < BIGBUF) str[j++] = ch;
	    }
	    ungetc(ch, fp);
	    str[j] = '\0';
	    /* Need to recognize nil specially here. */
	    if (strcmp("nil", str) == 0) {
		return lispnil;
	    } else {
		return intern_symbol(str);
	    }
	}
    }
    return lispeof;
}

/* Read a sequence of expressions terminated by a closing paren. */

Obj *
read_list(fp)
FILE *fp;
{
    Obj *thecar, *thecdr;

    thecar = read_form_aux(fp);
    if (thecar == (Obj *) closingparen) {  /* yechhy casting */
	return lispnil;
    } else if (thecar == lispeof) {
	init_warning("missing a closing parenthesis");
	return lispeof;
    } else {
	thecdr = read_list(fp);
	if (thecdr == lispeof) return lispeof;
	return (cons(thecar, thecdr));
    }
}

/* The usual list length function. */

length(list)
Obj *list;
{
    int rslt = 0;

    while (list != lispnil) {
	list = cdr(list);
	rslt++;
    }
    return rslt;
}


/* Basic allocation routines. */

Obj *
new_string(str)
char *str;
{
    Obj *new = newobj();

    new->type = STRING;
    new->v.str = str;
    return new;
}

Obj *
new_number(num)
int num;
{
    Obj *new = newobj();

    new->type = NUMBER;
    new->v.num = num;
    return new;
}

Obj *
new_utype(u)
int u;
{
    Obj *new = newobj();

    new->type = UTYPE;
    new->v.num = u;
    return new;
}

Obj *
new_mtype(r)
int r;
{
    Obj *new = newobj();

    new->type = MTYPE;
    new->v.num = r;
    return new;
}

Obj *
new_ttype(t)
int t;
{
    Obj *new = newobj();

    new->type = TTYPE;
    new->v.num = t;
    return new;
}

Obj *
new_method()
{
    Obj *new = newobj();

    new->type = METHOD;
    return new;
}

Obj *
cons(x, y)
Obj *x, *y;
{
    Obj *new = newobj();

    new->type = CONS;  
    new->v.cons.car = x;  new->v.cons.cdr = y;
    return new;
}

/* The usual suspects. */

Obj *
car(x)
Obj *x;
{
    if (x->type == CONS || x->type == NIL) {
	return x->v.cons.car;
    } else {
	char buf[BUFSIZE];

	sprintlisp(buf, x);
	init_warning("Car of non-list `%s' being taken, returning nil", buf);
	return lispnil;
    }
}

Obj *cdr(x)
Obj *x;
{
    if (x->type == CONS || x->type == NIL) {
	return x->v.cons.cdr;
    } else {
	char buf[BUFSIZE];

	sprintlisp(buf, x);
	init_warning("Cdr of non-list `%s' being taken, returning nil", buf);
	return lispnil;
    }
}

Obj *cadr(x) Obj *x; {  return car(cdr(x));  }

Obj *cddr(x) Obj *x; {  return cdr(cdr(x));  }

Obj *caddr(x) Obj *x; {  return car(cdr(cdr(x)));  }

/* Return the string out of both strings and symbols. */

char *
c_string(x)
Obj *x;
{
    return (symbolp(x) ? symboltable[x->v.sym.symnum].name : x->v.str);
}

c_number(x)
Obj *x;
{
    return x->v.num;
}

Obj *
intern_symbol(str)
char *str;
{
    int n;

    if (numsymbols >= maxsymbols) {
	run_error("can't intern \"%s\"; no more symbols allowed!", str);
    }
    if ((n = lookup_string(str)) >= 0) {
	return symboltable[n].symbol;
    } else {
	Obj *new = newobj();

	new->type = SYMBOL;
	new->v.sym.symnum = numsymbols;
	/* Declare a newly created symbol to be unbound. */
	new->v.sym.value = (Obj *) -12345;
	symboltable[numsymbols].name = copy_string(str);
	symboltable[numsymbols].symbol = new;
	symboltable[numsymbols].setter = NULL;
	numsymbols++;
	return new;
    }
}

/* Given a string, try to find a symbol with that as its name. */

lookup_string(str)
char *str;
{
    int i;

    for (i = 0; i < numsymbols; ++i) {
	if (strcmp(symboltable[i].name, str) == 0) return i;
    }
    return (-1);
}

Obj *
symbol_value(sym)
Obj *sym;
{
    Obj *val = sym->v.sym.value;

    if (val == (Obj *) -12345) {
	run_warning("Unbound symbol `%s', substituting nil", c_string(sym));
	val = lispnil;
    }
    return val;
}

Obj *
setq(sym, x)
Obj *sym, *x;
{
    /* what if not a symbol? */
    sym->v.sym.value = x;
    return x;
}

makunbound(sym)
Obj *sym;
{
    sym->v.sym.value = (Obj *) -12345;
}

/* Create a naming method. */

Obj *
make_namer(code)
Obj *code;
{
    Obj *namer = (Obj *) new_method();

    namer->v.meth.code = code;
    namer->v.meth.len = length(code);  /* efficiency hack */
    return namer;
}

numberp(x) Obj *x; { return (x->type == NUMBER); }

stringp(x) Obj *x; { return (x->type == STRING); }

symbolp(x) Obj *x; { return (x->type == SYMBOL); }

consp(x) Obj *x; { return (x->type == CONS); }

utypep(x) Obj *x; { return (x->type == UTYPE); }

mtypep(x) Obj *x; { return (x->type == MTYPE); }

ttypep(x) Obj *x; { return (x->type == TTYPE); }

methodp(x) Obj *x; { return (x->type == METHOD); }

boundp(sym) Obj *sym; { return (sym->v.sym.value != ((Obj *) -12345)); }

/* General structural equality test.  Assumes that it is not getting
   passed any circular structures. */

equal(x, y)
Obj *x, *y;
{
    if (x->type != y->type) return FALSE;
    switch (x->type) {
      case NUMBER:
	return (c_number(x) == c_number(y));
      case STRING:
	return (strcmp(c_string(x), c_string(y)) == 0);
      case SYMBOL:
	return (strcmp(c_string(x), c_string(y)) == 0);
      case CONS:
	return (equal(car(x), car(y)) && equal(cdr(x), cdr(y)));
      case UTYPE:
      case MTYPE:
      case TTYPE:
	return (c_number(x) == c_number(y));
      default:
	/* Maybe should case_panic instead. */
	return FALSE;
    }
}

member(x, lis)
Obj *x, *lis;
{
    if (lis == lispnil) {
	return FALSE;
    } else if (!consp(lis)) {
	/* should probably be an error of some sort */
	return FALSE;
    } else if (equal(x, car(lis))) {
	return TRUE;
    } else {
	return member(x, cdr(lis));
    }
}

/* Return the nth element of a list. */

Obj *
elt(lis, n)
Obj *lis;
int n;
{
    while (n-- > 0) {
	lis = cdr(lis);
    }
    return car(lis);
}

#ifdef USE_CONSOLE
/* Need this for reasonable printing. */

printlisp(obj)
Obj *obj;
{
    fprintlisp(stdout, obj);
}
#endif

fprintlisp(fp, obj)
FILE *fp;
Obj *obj;
{
    int i;

    switch (obj->type) {
      case NIL:
	fprintf(fp, "nil");
	break;
      case NUMBER:
	fprintf(fp, "%d", obj->v.num);
	break;
      case STRING:
	/* (should print escape chars if needed) */
	fprintf(fp, "\"%s\"", obj->v.str);
	break;
      case SYMBOL:
	/* (should print escape chars if needed) */
	fprintf(fp, "%s", c_string(obj));
	break;
      case CONS:
	fprintf(fp, "(");
	fprintlisp(fp, car(obj));
	/* No dotted pairs allowed in our version of Lisp. */
	fprint_list(fp, cdr(obj));
	break;
      case UTYPE:
	fprintf(fp, "u#%d", obj->v.num);
	break;
      case MTYPE:
	fprintf(fp, "r#%d", obj->v.num);
	break;
      case TTYPE:
	fprintf(fp, "t#%d", obj->v.num);
	break;
      default:
	fprintf(fp, "<???_@_0x%x>", (int) obj);
	break;
    }
}

fprint_list(fp, obj)
FILE *fp;
Obj *obj;
{
    Obj *tmp;

    for (tmp = obj; tmp != lispnil; tmp = cdr(tmp)) {
	fprintf(fp, " ");
	fprintlisp(fp, car(tmp));
    }	
    fprintf(fp, ")");
}

sprintlisp(buf, obj)
char *buf;
Obj *obj;
{
    int i;

    if (strlen(buf) + 10 > BUFSIZE) return;
    switch (obj->type) {
      case NIL:
	sprintf(buf, "nil");
	break;
      case NUMBER:
	sprintf(buf, "%d", obj->v.num);
	break;
      case STRING:
	/* (should print escape chars if needed) */
	sprintf(buf, "\"%s\"", obj->v.str);
	break;
      case SYMBOL:
	/* (should print escape chars if needed) */
	sprintf(buf, "%s", c_string(obj));
	break;
      case CONS:
	sprintf(buf, "(");
	sprintlisp(buf+strlen(buf), car(obj));
	/* No dotted pairs allowed in our version of Lisp. */
	sprint_list(buf+strlen(buf), cdr(obj));
	break;
      case UTYPE:
	sprintf(buf, "u#%d", obj->v.num);
	break;
      case MTYPE:
	sprintf(buf, "r#%d", obj->v.num);
	break;
      case TTYPE:
	sprintf(buf, "t#%d", obj->v.num);
	break;
      default:
	sprintf(buf, "<???_@_0x%x>", (int) obj);
	break;
    }
}

sprint_list(buf, obj)
char *buf;
Obj *obj;
{
    Obj *tmp;

    buf[0] = '\0';
    for (tmp = obj; tmp != lispnil; tmp = cdr(tmp)) {
	if (strlen(buf) + 10 > BUFSIZE) return;
	strcat(buf, " ");
	sprintlisp(buf+strlen(buf), car(tmp));
    }	
    sprintf(buf+strlen(buf), ")");
}

#ifdef DEBUGGING
/* For debugging only. */

dl(x)
Obj *x;
{
    fprintlisp(stderr, x);
}
#endif /* DEBUGGING */

/* Our version of evaluation derefs symbols and evals through lists,
   unless the list car is a "special form". */

Obj *
eval(x)
Obj *x;
{
    Obj *specform;

    switch (x->type) {
      case SYMBOL:
	return eval_symbol(x);
      case CONS:
	specform = car(x);
	if (symbolp(specform)) {
	    if (strcmp("append", c_string(specform)) == 0) {
		/* should flatten/append lists */
	    } else {
	    }
	}
	return eval_list(x);
      default:
	return x;
	break;
    }
}

/* There are some special symbols for whom evaluation is an unusual
   process. [urt]* are constructed only when needed. */

Obj *
eval_symbol(sym)
Obj *sym;
{
    char *name = c_string(sym);
    int u, m, t;
    Obj *typelist;

    if (boundp(sym)) {
	return symbol_value(sym);
    } else if (strcmp(name, "u*") == 0) {
	typelist = lispnil;
	for (u = numutypes-1; u >= 0; --u) {
	    typelist = cons(new_utype(u), typelist);
	}
	return setq(sym, typelist);
    } else if (strcmp(name, "m*") == 0) {
	typelist = lispnil;
	for (m = nummtypes-1; m >= 0; --m) {
	    typelist = cons(new_mtype(m), typelist);
	}
	return setq(sym, typelist);
    } else if (strcmp(name, "t*") == 0) {
	typelist = lispnil;
	for (t = numttypes-1; t >= 0; --t) {
	    typelist = cons(new_ttype(t), typelist);
	}
	return setq(sym, typelist);
    } else {
	run_warning("`%s' is unbound, returning self\n", c_string(sym));
	/* kind of a hack */
	return sym;
    }
}

/* List evaluation just blasts straight through the list. */

Obj *
eval_list(lis)
Obj *lis;
{
    if (lis == lispnil) {
	return lispnil;
    } else {
	return cons(eval(car(lis)), eval_list(cdr(lis)));
    }
}
