/* Copyright (c) 1987, 1988, 1991, 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* This file manages the scorekeepers in a game. */

#include "conq.h"

Obj *eval_sk_form();

/* The head of the list of scorekeepers. */

Scorekeeper *scorekeepers = NULL;

int numscorekeepers = 0;

Side *winner = NULL;    /* the winning side (allies are also winners) */

int numscores = 0;

/* Clear out any possible scorekeepers. */

init_scorekeepers()
{
    scorekeepers = NULL;
    numscorekeepers = 0;
}

Scorekeeper *
create_scorekeeper()
{
    Scorekeeper *sk = (Scorekeeper *) xmalloc(sizeof(Scorekeeper));

    sk->id = 0;
    sk->name = NULL;
    sk->when = lispnil;
    sk->who = lispnil;
    sk->knownto = lispnil;
    sk->trigger = lispnil;
    sk->body = lispnil;
    sk->record = lispnil;
    sk->notes = lispnil;
    sk->scorenum = numscores++;
    sk->initial = 0;
    sk->triggered = FALSE;
    sk->next = scorekeepers;
    scorekeepers = sk;
    ++numscorekeepers;
    return sk;
}

/* Allocate and fill in the initial score records for each side. */

init_scores()
{
    Side *side;
    Scorekeeper *sk;

    for_all_sides(side) {
	if (numscores > 0) {
	    side->scores = (short *) xmalloc(numscores * sizeof(short));
	    for_all_scorekeepers(sk) {
		side->scores[sk->scorenum] = sk->initial;
	    }
	}
    }
}

/* Test all the scorekeepers that should be run immediately before any
   side moves anything. */

check_turn_start_scores()
{
    Side *side;
    Scorekeeper *sk;

    if (keeping_score()) {
	Dprintf("Checking start-of-turn scorekeepers\n");
	for_all_scorekeepers(sk) {
	    if (symbolp(sk->when)
	        && strcmp(c_string(sk->when), "before-turn") == 0) {
		for_all_sides(side) {
		    if (scorekeeper_applicable(side, sk)) {
			run_scorekeeper(side, sk); 
		    }
		}
	    }
	}
    }
}

/* Test all the scorekeepers that should be run only at the end of a turn. */

check_turn_end_scores()
{
    Side *side;
    Scorekeeper *sk;

    if (keeping_score()) {
	Dprintf("Checking end-of-turn scorekeepers\n");
	for_all_scorekeepers(sk) {
	    if (sk->when == lispnil
	    	|| (symbolp(sk->when)
	    		&& strcmp(c_string(sk->when), "after-turn") == 0)) {
		for_all_sides(side) {
		    if (scorekeeper_applicable(side, sk)) {
			run_scorekeeper(side, sk); 
		    }
		}
	    }
	}
    }
}

check_after_action_scores()
{
	Side *side;
	Scorekeeper *sk;

	if (checking_after_actions()) {
		for_all_scorekeepers(sk) {
	    	if (symbolp(sk->when)
	    	    && strcmp(c_string(sk->when), "after-action") == 0) {
	    		if (sk->trigger == lispnil || sk->triggered) {
				for_all_sides(side) {
		    			if (scorekeeper_applicable(side, sk)) {
						run_scorekeeper(side, sk); 
		    			}
		    		}
		    	}
		    }
		}
	}
}

checking_after_actions()
{
	Scorekeeper *sk;

	if (!keeping_score()) return FALSE;
	for_all_scorekeepers(sk) {
		if (symbolp(sk->when) && strcmp(c_string(sk->when), "after-action") == 0) {
			return TRUE;
		}
	}
	return FALSE;
}

check_after_event_scores()
{
	Side *side;
	Scorekeeper *sk;

	if (checking_after_events()) {
		for_all_scorekeepers(sk) {
	    	if (symbolp(sk->when)
	    	    && strcmp(c_string(sk->when), "after-event") == 0) {
	    		if (sk->trigger == lispnil || sk->triggered) {
				for_all_sides(side) {
		    			if (scorekeeper_applicable(side, sk)) {
						run_scorekeeper(side, sk); 
		    			}
		    		}
		    	}
		    }
		}
	}
}

checking_after_events()
{
	Scorekeeper *sk;

	if (!keeping_score()) return FALSE;
	for_all_scorekeepers(sk) {
		if (symbolp(sk->when) && strcmp(c_string(sk->when), "after-event") == 0) {
			return TRUE;
		}
	}
	return FALSE;
}

/* This is true when the given side should be tested against the given sk. */

scorekeeper_applicable(side, sk)
Side *side;
Scorekeeper *sk;
{
    return (side->ingame
    		&& 1 /* sk applies to the side */
    		&& (sk->trigger == lispnil || sk->triggered)
    		);
}

run_scorekeeper(side, sk)
Side *side;
Scorekeeper *sk;
{
    eval_sk_form(side, sk, sk->body);
}

Obj *
eval_sk_form(side, sk, form)
Side *side;
Scorekeeper *sk;
Obj *form;
{
    int incr = 1;
    char *formtype;
    Obj *test;

    if (symbolp(form)) {
        if (strcmp(c_string(form), "last-side-wins") == 0) {
            eval_sk_last_side_wins(side, sk);
        }
    } else if (consp(form) && symbolp(car(form))) {
	formtype = c_string(car(form));
	switch (keyword_value(formtype)) {
	  case K_IF:
	    test = cadr(form);
	    if (eval_sk_test(side, sk, test)) {
	    	eval_sk_form(side, sk, car(cddr(form)));
	    }
	    break;
	  case K_WIN:
	    side_wins(side);
	    break;
	  case K_LOSE:
	    side_loses(side);
	    break;
	  case K_ADD:
	    if (cdr(form) != lispnil) {
		incr = c_number(cadr(form));
	    }
	    side->scores[sk->scorenum] += incr;
	    break;
	  case K_SUBTRACT:
	    if (cdr(form) != lispnil) {
		incr = c_number(cadr(form));
	    }
	    side->scores[sk->scorenum] -= incr;
	    break;
	  default:
	    if (strcmp(c_string(formtype), "count") == 0) {
	    	int count = 0;
	    	
	    	return new_number(count);
	    }
	    run_warning("unknown form type `%s' in scorekeeper", formtype);
	}
    } else {
    }
    return lispnil;
}

eval_sk_test(side, sk, form)
Side *side;
Scorekeeper *sk;
Obj *form;
{
    char *formtype;

    if (consp(form)) {
	formtype = c_string(car(form));
	if (strcmp(formtype, "and") == 0) {
	} else if (strcmp(formtype, "or") == 0) {
	} else if (strcmp(formtype, "not") == 0) {
	} else if (strcmp(formtype, "=") == 0) {
		return (c_number(eval_sk_form(side, sk, cadr(form))) ==
			c_number(eval_sk_form(side, sk, car(cddr(form)))));
	} else if (strcmp(formtype, "<") == 0) {
		return (c_number(eval_sk_form(side, sk, cadr(form))) <
			c_number(eval_sk_form(side, sk, car(cddr(form)))));
	} else if (strcmp(formtype, "<=") == 0) {
		return (c_number(eval_sk_form(side, sk, cadr(form))) <=
			c_number(eval_sk_form(side, sk, car(cddr(form)))));
	} else if (strcmp(formtype, ">") == 0) {
		return (c_number(eval_sk_form(side, sk, cadr(form))) >
			c_number(eval_sk_form(side, sk, car(cddr(form)))));
	} else if (strcmp(formtype, ">=") == 0) {
		return (c_number(eval_sk_form(side, sk, cadr(form))) >=
			c_number(eval_sk_form(side, sk, car(cddr(form)))));
	} else {
	    /* (should warn about garbage) */
	    return FALSE;
	}
    } else if (symbolp(form)) {
    }
    return FALSE;
}

eval_sk_last_side_wins(side, sk)
Side *side;
Scorekeeper *sk;
{
    Side *side2, *winner;
    Unit *unit;
    int numleft = 0, numu = 0;

    if (numsides >= 2) {
	    for_all_sides(side2) {
	    numu = 0;
	    for_all_side_units(side2, unit) {
	    	if (in_play(unit)) ++numu;
	    		    }
	    if (numu == 0) {
		side_loses(side2);
	    } else if (side2->ingame) {
	    	++numleft;
	    	winner = side2;
	    }
	}
	if (numleft == 1) {
	    side_wins(winner);
	}
    }
}

num_sides_in_game()
{
    int rslt = 0;
    Side *side;

    for_all_sides(side) {
	if (side->ingame) ++rslt;
    }
    return rslt;
}

side_wins(side, sk)
Side *side;
Scorekeeper *sk;
{
    side->status = 1;
    record_event(H_SIDE_WON, -1, side);
    remove_side_from_game(side);
}

side_loses(side, sk)
Side *side;
Scorekeeper *sk;
{
    int ux, uy;
    Unit *unit;
    Side *side2;

#if 0
    /* These msgs should come from game design via scorekeeper? */
    while (side->unithead->next != side->unithead) {
	unit = side->unithead->next;
	if (alive(unit)) {
	    ux = unit->x;  uy = unit->y;
	    unit_changes_side(unit, victor, CAPTURE, SURRENDER);
	    all_see_hex(ux, uy);
	} else {
	    /* units could die instead of changing sides.   Inefficient if */
	    /* it happens often, but sides don't resign often.  We need to */
	    /* flush the dead units to make sure that the side has no */
	    /* units.  Easier this way, since unit_changes_side could get */
	    /* a lot of occupants, so we can't just step through the list. */
	    /* sts: I don't believe this, but am not sure enough to hack. */
	    flush_dead_units();
	}
    }
#endif
    /* Add the mark of shame itself. */
    side->status = -1;
    record_event(H_SIDE_LOST, -1, side);
    remove_side_from_game(side);
}

/* Add scorefile handling here. */

/* (only done if some scorekeeper says how?) */

record_into_scorefile()
{
}
